MobSpells = LibStub("AceAddon-3.0"):NewAddon("MobSpells")
local AceGUI = LibStub("AceGUI-3.0")

LibStub("AceAddon-3.0"):EmbedLibraries(
	MobSpells,
	"AceEvent-3.0",
	"AceConsole-3.0"
)

local defaults = {global = {mobs = {}}}
local mobs
local bit_band = bit.band
local frame

-- Table recycling
local new, newHash, newSet, del
do
	local list = setmetatable({}, {__mode='k'})
	function new(...)
		local t = next(list)
		if t then
			list[t] = nil
			for i = 1, select('#', ...) do
				t[i] = select(i, ...)
			end
			return t
		else
			return {...}
		end
	end
	function newHash(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...), 2 do
			t[select(i, ...)] = select(i+1, ...)
		end
		return t
	end
	function newSet(...)
		local t = next(list)
		if t then
			list[t] = nil
		else
			t = {}
		end	
		for i = 1, select('#', ...) do
			t[select(i, ...)] = true
		end
		return t
	end
	function del(t)
		setmetatable(t, nil)
		for k in pairs(t) do
			t[k] = nil
		end
		t[''] = true
		t[''] = nil
		list[t] = true
		return nil
	end
end

function MobSpells:OnInitialize()
	self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	self.db = LibStub("AceDB-3.0"):New("MobSpellsDB", defaults)

	mobs = self.db.global.mobs
	self:RegisterChatCommand("mobspells", "Show")
	self.searchIndex = self.searchIndex or "NPC"
end

local spell_events = {
	SPELL_DAMAGE = true, 
	SPELL_CAST_START = true, 
	SPELL_HEAL = true, 
	SPELL_MISSED = true, 
	SPELL_PERIODIC_MISSED = true,
	SPELL_PERIODIC_DAMAGE = true, 
	SPELL_PERIODIC_HEAL = true,
	SPELL_AURA_APPLIED = true
}
function MobSpells:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
	local timestamp, eventtype, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags = ...
	if not spell_events[eventtype] then return end

	--[[
	if eventtype == "SPELL_AURA_APPLIED" then
		for i = 1, select("#", ...) do
			local v = select(i, ...)
			ChatFrame1:AddMessage(("%s: %s"):format(i, tostring(v)))
		end
	end
	]]--
	local valid, guid, name
	if bit_band(srcFlags, COMBATLOG_OBJECT_CONTROL_NPC) == COMBATLOG_OBJECT_CONTROL_NPC and 
	   bit_band(srcFlags, COMBATLOG_OBJECT_TYPE_NPC) == COMBATLOG_OBJECT_TYPE_NPC then
		valid = true
		guid = srcGUID
		name = srcName
	elseif srcName == nil and bit_band(dstFlags, COMBATLOG_OBJECT_CONTROL_NPC) == COMBATLOG_OBJECT_CONTROL_NPC and
		bit_band(dstFlags, COMBATLOG_OBJECT_REACTION_HOSTILE) == COMBATLOG_OBJECT_REACTION_HOSTILE and 
		bit_band(dstFlags, COMBATLOG_OBJECT_TYPE_NPC) == COMBATLOG_OBJECT_TYPE_NPC then
		local auraType = select(12, ...)
		if auraType == AURA_TYPE_BUFF then
			valid = true
			guid = dstGUID
			name = dstName
		end
	end
	if valid then
		local NPCID = tonumber(guid:sub(-12,-7),16)
		local spellID, spellName, spellSchool = select(9, ...)
		local zone = GetZoneText()
		mobs[zone] = mobs[zone] or {}
		mobs[zone][NPCID] = mobs[zone][NPCID] or {name = name, spells = {}}
		mobs[zone][NPCID].spells[spellID] = spellSchool
	end
end

function showSpellTooltip()
	GameTooltip:ClearLines();
	GameTooltip:SetOwner(this, "ANCHOR_RIGHT");
	GameTooltip:SetHyperlink(("spell:%s"):format(this.spellID))
	GameTooltip:AddLine(" ")
	GameTooltip:AddLine(("|cffffffffSpell ID #|h|cffffff00%s|h"):format(this.spellID))
	GameTooltip:Show()
	this.highlight:Show()
end

function hideTooltip()
	GameTooltip:Hide()
	this.highlight:Hide()
end

function addLink()
	if IsShiftKeyDown() and ChatFrameEditBox:IsShown() then
		ChatFrameEditBox:Insert(GetSpellLink(this.spellID));
	end
end

function SelectSpell(widget, event, value)
	widget:ReleaseChildren()

	local zone, mobID = ("\001"):split(value)
	if not mobID then return end
	local zoneData = MobSpells.db.global.mobs[zone]
	local data = zoneData[tonumber(mobID)]
	
	local sf = AceGUI:Create("ScrollFrame")
	sf:SetLayout("Flow")
	
	local heading1 = AceGUI:Create("Heading")
	heading1:SetText(data.name)
	heading1.width = "fill"
	sf:AddChild(heading1)
	
	for v, school in pairs(data.spells) do	
		local name, rank, icon = GetSpellInfo(v)
		local label = AceGUI:Create("Label")
		if rank and strlen(rank) > 0 then
			label:SetText((" [%s (%s)]"):format(name, rank))
		else
			label:SetText((" [%s]"):format(name))
		end
		local colors = COMBATLOG_DEFAULT_COLORS.schoolColoring[school] or COMBATLOG_DEFAULT_COLORS.schoolColoring.default
		label.label:SetTextColor(colors.r, colors.g, colors.b, colors.a)
		label.width = "fill"
		local p, h, f = label.label:GetFont()
		label.label:SetFont("Fonts\\ARIALN.TTF", 15, f)
		label.image:SetHeight(22)
		label.image:SetWidth(22)
		label.frame:EnableMouse(true)
		label.frame.spellID = v
		label.frame:SetScript("OnEnter", showSpellTooltip)
		label.frame:SetScript("OnLeave", hideTooltip)
		label.frame:SetScript("OnMouseDown", addLink)
		label.frame.highlight = label.frame:CreateTexture("Highlight", "OVERLAY")
		label.frame.highlight:SetTexture("Interface\\Buttons\\UI-PlusButton-Hilight")
		label.frame.highlight:SetAllPoints()
		label.frame.highlight:SetBlendMode("ADD")
		label.frame.highlight:Hide()

		label:SetImage(icon)
		sf:AddChild(label)
	end
	
	widget:AddChild(sf)
end

function MobSpells:CreateFrame()
	if frame then return end
	
	local f = AceGUI:Create("Frame")
	f:SetTitle("Mob Spells")
	f:SetStatusText("")
	f:SetLayout("Flow")
	f:SetWidth(500)
	f:SetHeight(350)
	f.width = "fill"
	
	local dropdown = AceGUI:Create("Dropdown")
	dropdown:SetText("NPCs")
	dropdown:SetLabel("Search")
	dropdown.list = {NPC = "NPCs", SPELL = "Spells"}
	dropdown:SetWidth(100)
	dropdown:SetCallback("OnValueChanged",function(widget,event,value) MobSpells.searchIndex = value end )
	f:AddChild(dropdown)

	local edit = AceGUI:Create("EditBox")
	edit:SetText("")
	edit:SetWidth(200)
	edit:SetLabel("Keywords")
	edit:SetCallback("OnEnterPressed",function(widget,event,text) MobSpells:Filter(text) end )
	edit:SetCallback("OnTextChanged",function(widget,event,text) MobSpells:Filter(text) end )
	f:AddChild(edit)
	
	local t = AceGUI:Create("TreeGroup")
	t:SetLayout("Fill")
	t:SetCallback("OnGroupSelected", SelectSpell )
	t.width = "fill"
	t.height = "fill"
	
	f:AddChild(t)
	f.tree = t

	frame = f
end

local function clearTree(t)
	for k, v in pairs(t) do
		if type(v) == "table" then
			t[k] = clearTree(v)
		else
			t[k] = nil
		end	
	end
	return del(t)
end

local function comp(a, b)
	if a[1] > b[1] then
		return false
	elseif a[1] < b[1] then
		return true
	else
		if a[2] > b[2] then
			return false
		elseif a[2] < b[2] then
			return true
		else
			return false
		end
	end
end

local t = new()
function MobSpells:BuildTree()
	t = clearTree(t)
	t = new()
	local zones = new()
	for zone, mobList in pairs(mobs) do
		tinsert(zones, zone)
	end
	table.sort(zones)
	for _, zone in ipairs(zones) do
		local mobList = mobs[zone]
		local tzone = newHash()
		tzone.text = zone
		tzone.value = zone
		local children = new()
		tzone.children = children
		local mobNames = new()
		for mobID, data in pairs(mobList) do
			tinsert(mobNames, new(data.name, mobID))
		end
		table.sort(mobNames, comp)
		for _, t in ipairs(mobNames) do
			mobID = t[2]
			data = mobList[mobID]
			mob = new()
			mob.text = data.name
			mob.value = mobID
			tinsert(children, mob)
		end
		for i = 1, #mobNames do
			mobNames[i] = del(mobNames[i])
		end
		mobNames = del(mobNames)
		tinsert(t, tzone)
	end	
	zones = del(zones)
	return t
end

function MobSpells:Filter(text)
	local t = self:BuildTree()
	-- Iterate zones
	local zOffset = 0
	for i = 1, #t do
		v = t[i-zOffset]
		local offset = 0
		for j = 1, #v.children do
			local child = v.children[j-offset]
			local useMob = false
			if strlen(text) == 0 then
				useMob = true
			end
			if self.searchIndex == "NPC" and child.text:lower():match(text:lower()) then
				useMob = true
			elseif self.searchIndex == "SPELL" then
				for k, v in pairs(self.db.global.mobs[v.value][child.value].spells) do
					local name = GetSpellInfo(k)
					if name:lower():match(text:lower()) then
						useMob = true
						break
					end
				end
			end
			if not useMob then
				tremove(v.children, j - offset)
				offset = offset + 1
			end
		end
		if #v.children == 0 then
			tremove(t, i - zOffset)
			zOffset = zOffset + 1
		end
	end
	frame.tree:SetTree(t)
end

function MobSpells:Show()
	self:CreateFrame()
	frame.tree:SetTree(self:BuildTree())
	frame:Show()
end
